/*
 * AVR_Bus.c
 *
 * Created: 10.01.2012 12:33:05
 *  Author: S.Schwarzbauer
 */ 

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "myConfig.h"
#include "SUART/suart_bus.h"



#ifdef BUS_MASTER
#include "UART/uart.h"
#include "OneWire/onewire.h"
#include "OneWire/ds18x20.h"

#define MAXSENSORS 5
#define CMD_TSENSOR_TEMP 0x30
#define TEMP_MEASSURE_DELAY 3000	// = seconds / 0,02

uint8_t gSensorIDs[MAXSENSORS][OW_ROMCODE_SIZE];
uint16_t TimerCounter;

#else
#include "irmpconfig.h"
#include "irmp.h"
#endif



void timer0_init (void)
{
	#ifdef BUS_MASTER
    TCCR0B = (1<<CS02) | (1<<CS00) ;		//divide by 1024
	TimerCounter = TEMP_MEASSURE_DELAY;
	#else
	TCCR0B = (1<<CS01) | (1<<CS00) ;		//divide by 64
	#endif
    TIMSK0 |= (1<<TOIE0);					//enable timer interrupt	
}

#ifdef BUS_MASTER
void uart_put(uint8_t command, uint8_t *data, uint8_t len)
{
	uint8_t i;
	uart_putc(BUS_SYNC_BYTE_1);
	uart_putc(BUS_SYNC_BYTE_2);
	uart_putc((uint8_t)(sizeof(command) + len));
	uart_putc(command);
	
	for (i=0; i < len; i++)
	{
		uart_putc(data[i]);
	}
	
	
}

static uint8_t search_sensors(void)
{
	uint8_t i;
	uint8_t id[OW_ROMCODE_SIZE];
	uint8_t diff, nSensors;
	
	ow_reset();

	nSensors = 0;
	
	diff = OW_SEARCH_FIRST;
	while ( diff != OW_LAST_DEVICE && nSensors < MAXSENSORS ) {
		DS18X20_find_sensor( &diff, &id[0] );
		
		if( diff == OW_PRESENCE_ERR ) {	
			break;
		}
		
		if( diff == OW_DATA_ERR ) {
			break;
		}
		
		for ( i=0; i < OW_ROMCODE_SIZE; i++ )
			gSensorIDs[nSensors][i] = id[i];
		
		nSensors++;
	}
	
	return nSensors;
}
#else

void Read_IR_Code_Table(void)
{
	uint8_t i;
	
	// clear first gloabl ir code table
	memset(&busirmpcodes, 0xFF, sizeof(busirmpcodes));	
	
	for (i=0;i<(sizeof(busirmpcodes) / sizeof(BUS_IRMP_DATA));i+=sizeof(BUS_IRMP_DATA))
	{
		eeprom_read_block(&busirmpcodes[i / sizeof(BUS_IRMP_DATA)], (uint8_t *)0 +i, sizeof(BUS_IRMP_DATA));
		
		
		if (busirmpcodes[i / sizeof(BUS_IRMP_DATA)].command_IR_data.protocol == 0xFF)
		{
			// no ir code stored here, exit reading eeprom
			break;
		} 
	}
}

int IR_Code_Match(IRMP_DATA *irmp_data)
{
	uint8_t i;
	int res = -1;
	
	for (i=0; i < (sizeof(busirmpcodes) / sizeof(BUS_IRMP_DATA)); i+=sizeof(BUS_IRMP_DATA))
	{
		if (busirmpcodes[i / sizeof(BUS_IRMP_DATA)].command_IR_data.protocol == 0xFF)
		{
			// no ir code stored here, exit reading eeprom
			break;
		}
		else
		{
			// compare if irmp code match
			if (memcmp(irmp_data, &busirmpcodes[i / sizeof(BUS_IRMP_DATA)].command_IR_data, sizeof(IRMP_DATA) - 1) == 0)	
			{
				res = i / sizeof(BUS_IRMP_DATA);
				break;				
			}
		}
	}
	
	return res;
}
#endif

ISR (TIMER0_OVF_vect)
{
	#ifdef BUS_MASTER
	TCNT0 = 40;	
	
	if (!TimerCounter)
	{
		uint8_t nSensors;
		uint8_t i;
		int16_t decicelsius;
		uint8_t buffer[OW_ROMCODE_SIZE + sizeof(decicelsius)];
		
		nSensors = search_sensors();
		
		if ( DS18X20_start_meas( DS18X20_POWER_EXTERN, NULL ) 
			== DS18X20_OK) {
			_delay_ms( DS18B20_TCONV_12BIT );
			for ( i = 0; i < nSensors; i++ ) {

				if ( DS18X20_read_decicelsius( &gSensorIDs[i][0], &decicelsius )
				     == DS18X20_OK ) {
					memcpy(&buffer[0], &gSensorIDs[i][0], OW_ROMCODE_SIZE);
					memcpy(&buffer[OW_ROMCODE_SIZE], &decicelsius, sizeof(decicelsius));
					uart_put(CMD_TSENSOR_TEMP, &buffer, sizeof(buffer));		
					// decicelsius
					// range from -550:-55.0C to 1250:+125.0C -> min. 6+1 chars														
				}
			}	
			
			}			

		TimerCounter = TEMP_MEASSURE_DELAY;
	}
	
	TimerCounter--;
	
	#else
	TCNT0 = -17;	
	irmp_ISR();	
	#endif
}

int AcceptNodeID(uint8_t ID)
{
	#ifdef BUS_MASTER
	if (ID == BUS_MASTER_NODE_ID)
	#else
	if ((ID == BUS_SLAVE_NODE_ID) || (ID == BUS_SLAVE_GLOBAL_ID))
	#endif
	{
		return TRUE;
	} else {
		return FALSE;
	}
}


int main(void)
{
	#ifdef BUS_MASTER
	
	BUS_IRMP_DATA irmp_data[3];
	uint16_t len, tmp_data;
	uint8_t buffer[255];
	uint8_t i;
	
	uint8_t nSensors;
		
	uart_init((UART_BAUD_SELECT((BAUD),F_CPU)));
    init_suart();	
	timer0_init();                                                          // initialize timer 0

	sei();	
	
    while(1)
    {
		
		tmp_data = uart_getc();
		
		if (((tmp_data & 0xFF00) >> 8) == 0x00 )
		{
			switch (uart_state)
			{
			case 0x00:
					if ((tmp_data & 0xFF) == BUS_SYNC_BYTE_1)
					{
						uart_state++;
					}
				break;
			case 0x01:
					if ((tmp_data & 0xFF) == BUS_SYNC_BYTE_2)
					{
						uart_state++;
					}
				break;		
			case 0x02:
					len = tmp_data;
					uart_state++;
				break;
			case 0x03:
					buffer[0] = tmp_data & 0xFF; 
					i = 1;
					while (i < len)
					{
						tmp_data = uart_getc();
				
						if (((tmp_data & 0xFF00) >> 8) == 0x00 )
						{
							buffer[i] = tmp_data & 0xFF;
							i++;
						}														
					}					
					uart_state = 0;
					
					// do command handling
					switch (buffer[0])
					{
					case CMD_WRITE_IRMP_CODES:
			
						SendBytes(BUS_SLAVE_NODE_ID, CMD_WRITE_IRMP_CODES, &buffer[1], len);
						break;
					}					
				break;
			}			
		}
    }
	
	#else
	
	IRMP_DATA irmp_data;
	uint8_t RepeatCounter = 0;
	uint8_t MinRepeats = 5;
	int Return_CMD_Index;
	
	/* first set all pins to input */
    PORTB = 0xFF;   													/* activate all pull-ups */
    DDRB = 0;       													/* all pins input */
    PORTC = 0xFF;   													/* activate all pull-ups */
    DDRC = 0;       													/* all pins input */
    PORTD = 0xFF;   													/* activate all pull-ups */
    DDRD = 0;    														/* all pins input */
	
	uart_init((UART_BAUD_SELECT((BAUD),F_CPU)));
    init_suart();
	
	irmp_init();                                                            // initialize irmp
	timer0_init();                                                          // initialize timer 0
	
	
	// clear variabels
	Read_IR_Code_Table();
		
	sei();
	
    while(1)
    {
		if (irmp_get_data (&irmp_data))
		{
			Return_CMD_Index = IR_Code_Match(&irmp_data);
			
			if (Return_CMD_Index >= 0x00)
			{
				if (!(irmp_data.flags & IRMP_FLAG_REPETITION))				// first check if the IR command is repeated
				{
					RepeatCounter = 0;										// new command received, reset RepeatCounter
				}																	
				else
				{	
					RepeatCounter++;										// still the same command, inc RepeatCounter
				}		
			
				if ((RepeatCounter == 0) | ( RepeatCounter >= MinRepeats))	// only send interrupt if first command, or Repeatcounter >= MinRepeats
				{
					if (RepeatCounter >= MinRepeats)	
						RepeatCounter = MinRepeats;							// fix overflow for long key push

					SendBytes(BUS_MASTER_NODE_ID, CMD_IRMP_DECODED_CMD, (uint8_t*)(&busirmpcodes[Return_CMD_Index].bus_command), sizeof(uint8_t));
				}					
			}									
		}	   
    }
	#endif	
}

void suart_rx_handler( uint16_t x )
{
	
	uint8_t i;	
	
	switch (suart_state)
	{
	case 0x00:
			if ((x & 0xFF) == BUS_SYNC_BYTE_1)
			{
				suart_state++;
			}
		break;
	case 0x01:
			if ((x & 0xFF) == BUS_SYNC_BYTE_2)
			{
				suart_state++;
			}
		break;		
	case 0x02:
			if (AcceptNodeID((x & 0xFF)))
			{
				suart_state++;
			} else
			{
				suart_state = 0;
			}		
		break;
	case 0x03:
			if (expected_data_len == 0)
			{
				// new data transfer
				expected_data_len = x & 0xFF;	
				memset(&bus_buffer, 0x00, sizeof(bus_buffer));
			}
			else
			{
				bus_buffer[receive_data_len] = x & 0xFF;
				receive_data_len++;
			}
			

			
			if (receive_data_len == expected_data_len)
			{
				// remove crc checksum len:
				receive_data_len--;
				
				if (CRC_Valid(&bus_buffer[0], receive_data_len, bus_buffer[receive_data_len]))
				{
					#ifdef BUS_MASTER

					switch (bus_buffer[0])
					{
					case CMD_IRMP_DECODED_CMD:
					
						uart_put(CMD_IRMP_DECODED_CMD, &bus_buffer[1], sizeof(uint8_t));						
						break;
						
					case CMD_READ_IRMP_CODES:									
								
						uart_put(CMD_READ_IRMP_CODES, &bus_buffer[1], receive_data_len);	
						break;							
					}			
					
					#else
					
					switch (bus_buffer[0])
					{
					case CMD_WRITE_IRMP_CODES:
					
							// struct of one code is:
							// one byte: bus command byte for this code
							// five bytes: IRMP code data without flag byte
							
							// remove cmd byte len:
							receive_data_len--;	
												
							// save all data in eeprom						
							eeprom_write_block(&bus_buffer[1], (uint8_t *)0, receive_data_len);	
							
							// clear rest of eeprom
							for (i=0;i<(sizeof(busirmpcodes)-receive_data_len); i++)
							{
								eeprom_write_byte((uint8_t *)0+receive_data_len+i, 0xFF);	
							}
							
							// re-read ir code table
							Read_IR_Code_Table();
						break;
						
					case CMD_READ_IRMP_CODES:
					
							for (i=0; i<(sizeof(busirmpcodes) / sizeof(BUS_IRMP_DATA));i+=sizeof(BUS_IRMP_DATA))
							{
								if (busirmpcodes[i / sizeof(BUS_IRMP_DATA)].command_IR_data.protocol == 0xFF)
								{
									SendBytes(BUS_MASTER_NODE_ID, CMD_READ_IRMP_CODES, (uint8_t*)(&busirmpcodes), i);
									break;
								}																							
							}
							
								
							
						break;							
					}					
					
					#endif		
				}			
				
				expected_data_len = 0;
				receive_data_len = 0;
				suart_state = 0;
			}
		break;			
	}	
}