/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2007 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/

#include "suart.h"
#include "suart_hw.h"

#include <stdint.h>


/** Software Universial Serial (Asynchronous) Receiver Transmitter
 * This is a communication module that can transmit multidirectional
 * data over one line. Thus it allows connecting many devices over a
 * single common line.
 * 
 * The Module uses Differential Manchester encoding on the line.
 * Hardware requirements are only a single pin that can be configured
 * as input, output and input with pull-up Resistor. The line should
 * have a much higher pull-down resistors (at least 10 times the resistance
 * as the (internal) pull-up resistor. On many links, there is no pull
 * down required at all. The module requires also a timer. The timer
 * must be programmed to wake up at a specific time.
 * 
 * All Hardware specific modifications must be done in the file "suart_hw.h"
 * before the code can be used.
 * 
 * The line is either high or low in idle mode. The last transmitter will
 * activate it's pull up resistor if the last line state was high, so the
 * line should keep it's level. In case, this is not possible (because of
 * termination for ex.), all devices should have an equal RXTIMEOUT.
 * 
 * A transition always starts with a "0", so each receiver can measure
 * the bitrate and adjust it's timers.
 * (a "0" is a transition of the line at the beginning of the
 * bit period and in the middle of the period whereas a "1" is no 
 * transition at the beginning of the bit period but a transition in the
 * middle of the bit time).
 * 
 * After the start bit ("0"), DATALEN bits are sent or received. The
 * DATALEN value must be equal in all devices connected to the bus.
 * 
 * After the last bit was transmitted or received, the received data is
 * given to the application by INPUT(x). All receiving devices will wait
 * for 3 bit periods after the last transition before they attempt to send 
 * something theirselfs. This allows the previously transmitting device
 * to resume transmission in a packet-oriented environment at the correct
 * time of 1 to 2 bit periods after the last transition.
 * 
 */

/* Variables for TX Module */
uint8_t suart_txbitnr = 0;
volatile uint8_t suart_txnextbitnr = 0;
uint16_t suart_txnextdata = 0;
uint16_t suart_txdata = 0;

/* Variables for RX Module */
uint8_t suart_rxbitnr;
uint16_t suart_rxperiod = 0;
uint16_t suart_rxlasttime = 0;
uint16_t suart_rxdata = 0;

void suart_init( void ) {
    /* initialize hardware */
    SUART_INIT_HW();
    SUART_DBG_INIT();
    /* release line */
    SUART_INPUT();
}



/** send one datum over the line
 * This may not be called from an interrupt or
 * with interrupts disabled! It could block
 * if the tx fifo is full (1 datum)
 */
void suart_putc( uint16_t data ) {
    while (suart_txnextbitnr != 0);         // wait until we are not transmitting any more...
    // data must be "left" aligned
    suart_txnextdata = (data << ((sizeof(data)*8) - SUART_DATALEN)); // startbit is not in here!
    // number of half bits plus the one startbit:
    suart_txnextbitnr = ((SUART_DATALEN + 1) << 1);
    // if the timer interrupt is currently disabled, enable it and let timer expire very
    // shortly (start transmission)
    SOFT_TIMER_INTERRUPT_IF_DISABLED();
}

// time is the time, the interrupt was issued (for exacter calculation of next time)
__attribute__((always_inline)) inline void suart_timer( uint16_t time ) {
    SUART_DBG_A;
    
    /* There are several reasons for this:
    a) We were transmitting and we have to send the next half bit
    b) We were receiving and the transmitting device crashed or did
       not keep timing specification -> rx timeout
    c) Bus is idle but rxlogic has received a start bit (drift). this timeout
       should reset the logic again - and maybe pull line to new level.
    d) line is and was idle, we have new data and should start to send.
    */
    
    //a)
    // Check for transmitting. This must be the first case, because when we are
    // transmitting, the other timeouts should not happen at all.
    if (suart_txbitnr != 0) {
        // send out next bit half.
        // advance bit pointer.
        suart_txbitnr--;
        if ((suart_txbitnr & 1) != 0) {
            // we are in the second half of a bit ... change line
            SUART_SWAP();
        } else if (suart_txbitnr > 0) {
            // there are still bits to send
            if ((suart_txdata & 0x8000) != 0) {
                // a "1" is indicated by no transition at the beginning of the bit time
            } else {
                // a "0" is indicated by a transition at the beginning of the bit
                SUART_SWAP();
            }
            suart_txdata = suart_txdata << 1;
        }
        // Check if this was the end of the last bit half.
        if (suart_txbitnr == 0) {
            // no more bits to send, make line idle with pullup if needed
            if (SUART_INPUT_LEVEL() != 0)
                SUART_PULLUP()
            else
                SUART_INPUT();
            // wait at half a transition
            // We must wait for half period so a new transition that is initiated by us
            // will not be interrupted by some other one's start. (see rxlogic comments)
            SUART_SET_TIMER( time + SUART_TXPERIOD / 2 );
        } else {
            // more bits to be sent, wait one period
            SUART_SET_TIMER( time + SUART_TXPERIOD );
        }
    }
    else {
        // b+c+d:
        if (suart_rxbitnr != 0) {
            SUART_DBG_NB;
            // we were receiving and the last bit transition is too long ago or
            // we have already received the last bit and the sender did not start
            // a new transition. thus we can reset the rx logic (data already given
            // to application).
            // if this was a drift, we don't have a valid rxperiod, yet. Then, we should wait
            // for until every slave got knowledge of the drift change. This can currently
            // not ensured by a simple delay since every receiving slave sets it's
            // drift-detecting timeout (RXTIMEOUT) different.
            // It can be ensured, if the first device, who's RXTIMEOUT elapses, starts a
            // transmission of one bit. After the bit is transmitted, the state is as if
            // the transmitting device crashed. But the Timeout in all slaves will be
            // the same, because the rxperiod was "broadcasted" by the device that detected
            // the drift.
            // we must only ensure that the RXTIMEOUT of all devices is larger than all
            // RXPERIODs of all devices. Else a very slow RXPERIOD will be detected as
            // RXTIMEOUT by a fast device.
            if (suart_rxperiod == 0) {
                // drift detected
                //TODO: start a transmission of 1 bit or reset the rxlogic and hope, all
                // listening devices get knowledge of the drift before some device that already
                // registred the drift starts sending. Can be ensured by setting RXTIMEOUT
                // appropriately
            } else {
                // either last bit received or sender crashed.
                // if last bit, rx_bitnr should be 1.
                if (suart_rxbitnr == 1) {
                    // we have received the last bit and there are already two periods over.
                    // we should wait another period until we start a transmission from our side.
                    SUART_SET_TIMER( time + suart_rxperiod );
                } else {
                    // sender crashed, delay another period (we have a valid rxperiod)
                    // the total sum of time waited after last transition is 4 half periods
                    // (1 from here and 3 from rx logic)
                    SUART_SET_TIMER( time + suart_rxperiod );
                }
            }
            // reset the rx logic.
            suart_rxbitnr = 0;
        }
        // d:
        else {
            // line was idle, we can transmit
            // is there a need for transmission?
            if (suart_txnextbitnr != 0) {
                suart_txdata = suart_txnextdata;
                suart_txbitnr = suart_txnextbitnr;
                suart_txnextbitnr = 0;
                // set line to the inverse of the current state as output
                // this is due to the start bit always being 0!
                if (SUART_INPUT_LEVEL() != 0) {
                    SUART_OUTPUT( 0 );
                } else {
                    SUART_OUTPUT( 1 );
                }
                // after the startbit, the bit counter is uneven and the next
                // timer interrupt will swap the line and finish the start bit
                // only then the first bit of txdata will be shifted out.
                // this allows to use 16 databits with txdata beeing 16 bits wide
                SUART_SET_TIMER( time + SUART_TXPERIOD );
            } else {
                // no need for transmission, no reception active
                // turn off timer interrupt, it is enable either by
                // the next transition (start bit) or by the putc routine
                SUART_DISABLE_TIMER(); 
            }
        }
    }
    SUART_DBG_NA;
}


// must be called when the level on the pin changes
__attribute__((always_inline)) inline void suart_pinchange( void ) {
    uint16_t time = SUART_TIME;      // get time of entry into ISR
    SUART_DBG_A;                // isr entry debugging
    
    /* There was a transition on the BUS.
    This could be because of 
    a) a drift (errornous transition, very slow)
    b) a start transition
    c) a fast transition in case of errors
    d) a regular bit transition
    a+b are quite the same, we don't know if it is a or b.
    */
    
    // a+b:
    if (suart_rxbitnr <= 1) {
        // line was idle
        // clear receive registers
        suart_rxperiod = 0;
        suart_rxdata = 0;
        // initialize number of bits, here also the start bit must be added,
        // it will also be shifted into the rxdata but "fall out" if 16 bits 
        // are specified
        suart_rxbitnr = ((SUART_DATALEN + 1) << 1);
        // disable pull up resistor in case we are not transmitting
        if (suart_txbitnr == 0) {
            SUART_INPUT();
            // set timeout of the next transition to RXTIMEOUT,
            // this specifies the max time a bit transition may need.
            // if time is bigger than RXTIMEOUT, the transition will be
            // considered a drift and will be caught in the timer ISR.
            // do this only if we are not transmitting (else it would set the bitrate!)
            SUART_SET_TIMER( time + SUART_RXTIMEOUT );
            // and turn on timer interrupt
            SUART_ENABLE_TIMER();
        }
        SUART_DBG_NB;
    }
    // c+d:
    else {
        // we are already receiving some bits...
        // if we don't know the transmitting bitrate yet, we had the first bit here.
        // since the first bit is always 0, we will receive a single half-bit as first
        // transition and can set rxperiod now:
        if (suart_rxperiod == 0) {
            suart_rxperiod = time - suart_rxlasttime;
        }
        // we now know the bitrate of the sender now, so we can differentiate between a
        // too fast transition (error, c), a zero transition or a one transition (d)
        if (time - suart_rxlasttime < suart_rxperiod / 2) {
            // to fast transition
            // TODO: Need something here. could be: ignore this and the next transition.
        } else {
            if (time - suart_rxlasttime > suart_rxperiod + (suart_rxperiod / 2)) {
                // transition was longer than a single half bit period -> 1
                suart_rxdata = (suart_rxdata << 1) | 1;
                suart_rxbitnr -= 2;     // this was 2 half bits long
                SUART_DBG_C;
            } else {
                // are we in the middle of the bit?
                if ((suart_rxbitnr & 1) == 0) {
                    // yes, copy to memory, transition was (normal) short -> 0
                    suart_rxdata = (suart_rxdata << 1) | 0;
                    SUART_DBG_NC;
                }
                suart_rxbitnr -= 1;     // this was 1 half bits long
            }
        }
        // if we have more bits to come, we can also give a maximum margin for the
        // next bit transition time. Thus allowing the timer to reset the rxlogic
        // in case of a too long delay.
        if (suart_rxbitnr > 1) {
            // we allow a 3 half bits time here for some tolrance
            // don't set the timer, if we are transmitting
            if (suart_txbitnr == 0)
                SUART_SET_TIMER( time + suart_rxperiod * 3);
        } else {
            // we have received the last transition. The next one will either come
            // as a drift or as a new start transition after at suart_rxperiod
            // thus we can savely reset the rxlogic after suart_rxperiod after this bit finished (->2*period)
            // if we had ourselves put this byte on the bus, and we want another
            // byte to be sent, we don't want someone else to come before us, so
            // we should really only wait until rxperiod. But the other way around, devices
            // that were receiving, should wait at least two half bits time to allow
            // the sending device to resume the transmission for packet oriented transfers
            if (suart_txbitnr != 0) {
                // we were sending, timeout was already set by tx timer interrupt (tx logic)
            } else {
                // we were receiving, set timeout to 2*rxperiod
                SUART_SET_TIMER( time + suart_rxperiod * 2 );
#if ((! defined SUART_LOCAL_ECHO) || (!SUART_LOCAL_ECHO))
                SUART_INPUT_HANDLER( suart_rxdata );
#endif
            }
            // in either case, we can savely handle the received value to the application, or ignore
            // it if no local echo specified and we were transmitting
#if ((defined SUART_LOCAL_ECHO) && (SUART_LOCAL_ECHO))
            SUART_INPUT_HANDLER( suart_rxdata );
#endif
            SUART_DBG_B;
        }
    }
    suart_rxlasttime = time;
    SUART_DBG_NA;
}


SUART_INTERRUPTS;
