/*

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Copyright (C) 2007 by Clemens Helfmeier,
    Clemens(at)Helfmeier[dot]de

*/

#ifndef SUART_HW_H
#define SUART_HW_H

#include <stdint.h>

/** API
 * ===================
 */

/** Input handler. All data that was received is given to
 * this routine. The routine should return as fast as possible,
 * it is called from inside an interrupt!
 * Data x is right aligned!
 */
#define SUART_INPUT_HANDLER(x)      suart_rx_handler(x)
void suart_rx_handler( uint16_t x );

/** Defines if data sent by this slave will also be processed
 * by the receiver. This could be used for error detection
 * (collision). The data returned to the application will
 * represent the data that was "received" on this device's
 * input pin. There is no guarantee for the remote devices
 * to have received the correct data.
 */
#define SUART_LOCAL_ECHO    0

/** Baud Rate Specification (TX Only!)
 * The Baud Rate is given in Timer increments.
 * real Frequency is 1/(TXPERIOD * T) for
 * "0"-only and 1/(2*TXPERIOD * T) for "1"-only
 * The Maximum baud rate for AVR is currently ~350 Prozessor
 * cycles (watch out for Prescaler).
 * The maximum baudrate that can be recieved is only defined
 * by the interrupt execution time and the timer interval:
 * Bmax is about 1/(2 * T), the lower the better ;-)
 */
#define SUART_TXPERIOD      (1000)
/** RX Timeout Specification 
 * This sets the minimum Baud Rate accepted by
 * this device.
 * Real minimum Frequency is 1/(RXTIMEOUT * T) for
 * "0"-only and 1/(2*RXTIMEOUT * T) for "1"-only
 * It's a good choice to set it to a value of
 * 2/3*TimerOverflowValue
 */
#define SUART_RXTIMEOUT     (65535*1/3)


/** Code Length
 * The device will operate with 8 bits of data on the bus
 * Any value between 1 and 16 work here.
 */
#define SUART_DATALEN         8


/** Hardware Control
 * ==================
 */
// AVR Specific
#include <avr/interrupt.h>
#define SUART_BIT           PB0
#define SUART_PORT          PORTB
#define SUART_PIN           PINB
#define SUART_DIR           DDRB

/** Initialization routines, called once on initialization
          * of suart module. This should setup the pin, the input interrupt
          * and the timer interrupt. All interrupts should be enabled.
          * Pin should be an input (high impedance).
 */
#define SUART_INIT_HW()       {                             \
    TCCR1A = 0;                                             \
    SUART_INPUT();                                          \
    if (SUART_INPUT_LEVEL() != 0) {                         \
        TCCR1B = (0<<CS11) | (1<<CS10); /* falling edge */  \
    } else {                                                \
        TCCR1B = (0<<CS11) | (1<<CS10) | (1<<ICES1);    /* rising edge */ \
    }                                                       \
    OCR1A = 0;                                              \
    TIMSK1 |= (1<<OCIE1A) | (1<<ICIE1);                      \
}


/** Return the current level of the line. Either 0 or
          * any other value
 */
#define SUART_INPUT_LEVEL()     (SUART_PIN & (1<<SUART_BIT))

/** Turn on our pull-up resistor on the line
 */
#define SUART_PULLUP()          { SUART_PORT |= (1<<SUART_BIT); SUART_DIR &= ~(1<<SUART_BIT); }

/** Select our pin as input (high impedance, no pullup)
 */
#define SUART_INPUT()           { SUART_PORT &= ~(1<<SUART_BIT); SUART_DIR &= ~(1<<SUART_BIT); }

/** Swap the level on the pin. This is only called if
          * the pin is an output
 */
#define SUART_SWAP()            { SUART_PORT ^= (1<<SUART_BIT); }

/** Set the pin as output and set level of line
 */
#define SUART_OUTPUT( level )   {           \
    if (level != 0)                         \
        {SUART_PORT |= (1<<SUART_BIT);}     \
    else                                    \
        {SUART_PORT &= ~(1<<SUART_BIT);}    \
    SUART_DIR |= (1<<SUART_BIT);            \
}

/** Interrupt routines.
          * These defines will be simply added at the end of the suart.c file
          * so any interrupt can be put into here.
          * suart_pinchange and suart_timer are both inlined!
 */
#define SUART_INTERRUPTS        \
                 ISR (SIG_INPUT_CAPTURE1) {              \
                 TCCR1B ^= (1<<ICES1);   /* swap edge */   \
                 suart_pinchange();                  \
}                                       \
                 ISR (SIG_OUTPUT_COMPARE1A) {            \
                 suart_timer( OCR1A );               \
}

/** Timer
          * =======
          * The timer must increment continuosly it's value. Overflows
          * do not matter, calculation of times in suart.c is done with
          * uint16_t variables, but also 8 bit timer should work if the 
          * TXPERIOD and RXTIMEOUT values do not exceed the range.
 */
/** Return the Current execution time, This is called at the
          * very beginning of the pinchange interrupt handler, it could
          * also represent a capture register that was taken when the
          * interrupt was generated. Using this capture technique will
          * increase accuracy.
 */
#define SUART_TIME          ICR1

/** Set the time when the next timer interrupt should occure
          * and clear any pending timer interrupt (important!)
 */
#define SUART_SET_TIMER(x) { OCR1A = x; TIFR1 = (1<<OCIE1A); }

/** Enable the timer interrupt
 */
#define SUART_ENABLE_TIMER()    { TIMSK1 |= (1<<OCIE1A); }

/** Disable the timer interrupt
 */
#define SUART_DISABLE_TIMER()   { TIMSK1 &= ~(1<<OCIE1A); }

/** If the timer interrupt is disabled, enable it and set a very
          * short timeout. Between test for disabled interrupt and setting of
          * timeout, there should not be any other interrupt inside the suart
          * module (especially pinchange), the "function" must be atomic.
 */
#define SOFT_TIMER_INTERRUPT_IF_DISABLED()   {   \
    cli();                                      /* atomic */            \
    if ((TIMSK1 & (1<<OCIE1A)) == 0) {           /* test for en */       \
        TIMSK1 |= (1<<OCIE1A);                   /* if not, enable */    \
        OCR1A = TCNT1 + 5;                      /* set timeout to the lowest acceptable value */ \
    }                                                                   \
    sei();                                      /* atomic end */        \
}

/** Debugging defines.
 */
#if 0
#include <avr/io.h>
#define SUART_DBG_A     PORTB |= (1<<0);
#define SUART_DBG_NA    PORTB &= ~(1<<0);
#define SUART_DBG_B     PORTB |= (1<<1);
#define SUART_DBG_NB    PORTB &= ~(1<<1);
#define SUART_DBG_C     PORTB |= (1<<2);
#define SUART_DBG_NC    PORTB &= ~(1<<2);
#define SUART_DBG_INIT()  DDRB |= (1<<0) | (1<<1) | (1<<2);
#else
#define SUART_DBG_A
#define SUART_DBG_NA
#define SUART_DBG_B
#define SUART_DBG_NB
#define SUART_DBG_C
#define SUART_DBG_NC
#define SUART_DBG_INIT()
#endif

#endif
